"""pygeoprocessing: geoprocessing routines for GIS.

__init__ module imports all the geoprocessing functions into this namespace.
"""
import logging
import types

try:
    from importlib.metadata import PackageNotFoundError
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version
    from importlib_metadata import PackageNotFoundError

from . import geoprocessing
from .geoprocessing import _assert_is_valid_pixel_size
from .geoprocessing import align_and_resize_raster_stack
from .geoprocessing import align_bbox
from .geoprocessing import array_equals_nodata
from .geoprocessing import build_overviews
from .geoprocessing import calculate_disjoint_polygon_set
from .geoprocessing import choose_dtype
from .geoprocessing import choose_nodata
from .geoprocessing import convolve_2d
from .geoprocessing import create_raster_from_bounding_box
from .geoprocessing import create_raster_from_vector_extents
from .geoprocessing import distance_transform_edt
from .geoprocessing import get_gis_type
from .geoprocessing import get_raster_info
from .geoprocessing import get_vector_info
from .geoprocessing import interpolate_points
from .geoprocessing import iterblocks
from .geoprocessing import mask_raster
from .geoprocessing import merge_bounding_box_list
from .geoprocessing import new_raster_from_base
from .geoprocessing import numpy_array_to_raster
from .geoprocessing import raster_calculator
from .geoprocessing import raster_map
from .geoprocessing import raster_reduce
from .geoprocessing import raster_to_numpy_array
from .geoprocessing import rasterize
from .geoprocessing import ReclassificationMissingValuesError
from .geoprocessing import reclassify_raster
from .geoprocessing import reproject_vector
from .geoprocessing import shapely_geometry_to_vector
from .geoprocessing import stitch_rasters
from .geoprocessing import transform_bounding_box
from .geoprocessing import warp_raster
from .geoprocessing import zonal_statistics
from .geoprocessing_core import calculate_slope
from .geoprocessing_core import raster_band_percentile
from .slurm_utils import log_warning_if_gdal_will_exhaust_slurm_memory
from .utils import GDALUseExceptions
from .utils import gdal_use_exceptions

try:
    __version__ = version('pygeoprocessing')
except PackageNotFoundError:
    # package is not installed
    pass


# Programmatically defining __all__ based on what's been imported.
# Thus, the imports are the source of truth for __all__.
__all__ = ('calculate_slope', 'raster_band_percentile',
           'ReclassificationMissingValuesError')
exclude_set = {'log_warning_if_gdal_will_exhaust_slurm_memory'}
for attrname in [k for k in locals().keys()]:
    try:
        if (isinstance(getattr(geoprocessing, attrname), types.FunctionType)
                and attrname not in exclude_set):
            __all__ += (attrname,)
    except AttributeError:
        pass

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())  # silence logging by default

# these are bit masks for the known PyGeoprocessing types
UNKNOWN_TYPE = 0
RASTER_TYPE = 1
VECTOR_TYPE = 2

# Check GDAL's cache max vs SLURM memory if we're on slurm.
log_warning_if_gdal_will_exhaust_slurm_memory()



from pygeoprocessing.routing.helper_functions import extract_streams_d8
from pygeoprocessing.routing.routing import calculate_subwatershed_boundary
from pygeoprocessing.routing.routing import detect_lowest_drain_and_sink
from pygeoprocessing.routing.routing import detect_outlets
from pygeoprocessing.routing.routing import distance_to_channel_d8
from pygeoprocessing.routing.routing import distance_to_channel_mfd
from pygeoprocessing.routing.routing import extract_strahler_streams_d8
from pygeoprocessing.routing.routing import extract_streams_mfd
from pygeoprocessing.routing.routing import fill_pits
from pygeoprocessing.routing.routing import flow_accumulation_d8
from pygeoprocessing.routing.routing import flow_accumulation_mfd
from pygeoprocessing.routing.routing import flow_dir_d8
from pygeoprocessing.routing.routing import flow_dir_mfd
from pygeoprocessing.routing.watershed import delineate_watersheds_d8

__all__ = (
    'fill_pits',
    'fill_pits',
    'flow_dir_d8',
    'flow_accumulation_d8',
    'flow_dir_mfd',
    'flow_accumulation_mfd',
    'distance_to_channel_d8',
    'distance_to_channel_mfd',
    'extract_streams_d8',
    'extract_streams_mfd',
    'delineate_watersheds_d8',
    'detect_outlets',
    'extract_strahler_streams_d8',
    'calculate_subwatershed_boundary',
    'detect_lowest_drain_and_sink',
)



# coding=UTF-8
"""Multiprocessing implementation of raster_calculator."""
import collections
import errno
import multiprocessing
import os
import pprint
import signal
import sys
import time

import numpy
from osgeo import gdal

from ..geoprocessing import _is_raster_path_band_formatted
from ..geoprocessing import _LARGEST_ITERBLOCK
from ..geoprocessing import _LOGGING_PERIOD
from ..geoprocessing import _MAX_TIMEOUT
from ..geoprocessing import _VALID_GDAL_TYPES
from ..geoprocessing import geoprocessing_core
from ..geoprocessing import get_raster_info
from ..geoprocessing import iterblocks
from ..geoprocessing import LOGGER
from ..geoprocessing_core import DEFAULT_GTIFF_CREATION_TUPLE_OPTIONS
from ..utils import GDALUseExceptions

if sys.version_info >= (3, 8):
    import multiprocessing.shared_memory


def _block_success_handler(callback_state):
    """Used to update callback state after a successful block is complete.

    Updates the blocks complete and last time, if last_time has been >
    _LOGGING_PERIOD then dump a log.

    Args:
        callback_state (dict): contains the following keys
            'blocks_complete' -- number of raster calculator blocks processed
            'total_blocks' -- total number to process
            'last_time' -- last time.time() when a log was printed

    Returns:
        None
    """
    callback_state['blocks_complete'] += 1
    if time.time() - callback_state['last_time'] > _LOGGING_PERIOD:
        LOGGER.info(
            f"""raster_calculator {
                callback_state['blocks_complete']/
                callback_state['total_blocks']*100:.2f}% complete""")
        callback_state['last_time'] = time.time()


RasterPathBand = collections.namedtuple(
    'RasterPathBand', ['path', 'band_id'])


def _build_raster_calc_error_handler(pool):
    def _raster_calc_error_handler(exception):
        """Error handler for raster_calculator."""
        pool.terminate()
        raise Exception(
            f"error on raster calculator worker '{exception}'").with_traceback(
                exception.__traceback__)
    return _raster_calc_error_handler


def _raster_calculator_worker(
        block_offset_queue, base_canonical_arg_list, local_op,
        stats_worker_queue, nodata_target, target_raster_path, write_lock,
        processing_state, result_array_shared_memory):
    """Process a single block of an array for raster_calculation.

    Args:
        block_offset (dict): contains 'xoff', 'yoff', 'xwin_size', 'ywin_size'
            and can be used to pass directly to Band.ReadAsArray.
        base_canonical_arg_list (list): list of RasterPathBand, numpy arrays,
            or 'raw' objects to pass to the ``local_op``.
        local_op (function): callable that a function that must take in as
            many parameters as there are elements in
            ``base_canonical_arg_list``. Full description can be found in the
            public facing ``raster_calculator`` operation.
        stats_worker_queue (queue): pass a shared memory object ``local_op``
            result to queue if stats are being calculated. None otherwise.
        nodata_target (numeric or None): desired target raster nodata
        target_raster_path (str): path to target raster.
        write_lock (multiprocessing.Lock): Lock object used to coordinate
            writes to raster_path.
        processing_state (multiprocessing.Manager.dict): a global object to
            pass to ``__block_success_handler`` for this execution context.
        result_array_shared_memory (multiprocessing.shared_memory): If
            Python version >= 3.8, this is a shared
            memory object used to pass data to the stats worker process if
            required. Should be pre-allocated with enough data to hold the
            largest result from ``local_op`` given any ``block_offset`` from
            ``block_offset_queue``. None otherwise.

    Returns:
        None.

    """
    # read input blocks
    while True:
        block_offset = block_offset_queue.get()
        if block_offset is None:
            # indicates this worker should terminate
            return

        offset_list = (block_offset['yoff'], block_offset['xoff'])
        blocksize = (block_offset['win_ysize'], block_offset['win_xsize'])
        data_blocks = []
        for value in base_canonical_arg_list:
            if isinstance(value, RasterPathBand):
                with GDALUseExceptions():
                    raster = gdal.OpenEx(value.path, gdal.OF_RASTER)
                    band = raster.GetRasterBand(value.band_id)
                    data_blocks.append(band.ReadAsArray(**block_offset))
                    # I've encountered the following error when a gdal raster
                    # is corrupt, often from multiple threads writing to the
                    # same file. This helps to catch the error early rather
                    # than lead to confusing values of ``data_blocks`` later.
                    if not isinstance(data_blocks[-1], numpy.ndarray):
                        raise ValueError(
                            f"got a {data_blocks[-1]} when trying to read "
                            f"{band.GetDataset().GetFileList()} at "
                            f"{block_offset}, expected numpy.ndarray.")
                    raster = None
                    band = None
            elif isinstance(value, numpy.ndarray):
                # must be numpy array and all have been conditioned to be
                # 2d, so start with 0:1 slices and expand if possible
                slice_list = [slice(0, 1)] * 2
                tile_dims = list(blocksize)
                for dim_index in [0, 1]:
                    if value.shape[dim_index] > 1:
                        slice_list[dim_index] = slice(
                            offset_list[dim_index],
                            offset_list[dim_index] +
                            blocksize[dim_index],)
                        tile_dims[dim_index] = 1
                data_blocks.append(
                    numpy.tile(value[tuple(slice_list)], tile_dims))
            else:
                # must be a raw tuple
                data_blocks.append(value[0])

        target_block = local_op(*data_blocks)

        if (not isinstance(target_block, numpy.ndarray) or
                target_block.shape != blocksize):
            raise ValueError(
                "Expected `local_op` to return a numpy.ndarray of "
                "shape %s but got this instead: %s" % (
                    blocksize, target_block))

        with write_lock:
            with GDALUseExceptions():
                target_raster = gdal.OpenEx(
                    target_raster_path, gdal.OF_RASTER | gdal.GA_Update)
                target_band = target_raster.GetRasterBand(1)
                target_band.WriteArray(
                    target_block, yoff=block_offset['yoff'],
                    xoff=block_offset['xoff'])
                _block_success_handler(processing_state)
                target_band = None
                target_raster = None

        # send result to stats calculator
        if not stats_worker_queue:
            continue

        # Construct shared memory object to pass to stats worker
        if nodata_target is not None:
            target_block = target_block[target_block != nodata_target]
        target_block = target_block.astype(numpy.float64).flatten()

        if result_array_shared_memory:
            shared_memory_array = numpy.ndarray(
                target_block.shape, dtype=target_block.dtype,
                buffer=result_array_shared_memory.buf)
            shared_memory_array[:] = target_block[:]

            stats_worker_queue.put((
                shared_memory_array.shape, shared_memory_array.dtype,
                result_array_shared_memory))
        else:
            stats_worker_queue.put(target_block)


def _calculate_target_raster_size(
        raster_info_list, base_raster_path_band_const_list):
    """Determine the target raster size.

    Args:
        raster_info_list (list): list of raster info from
            ``base_raster_path_band_const_list``.
        base_raster_path_band_const_list (list/tuple): argument from
            raster_calculator.

    Returns:
        count of number of valid numpy array elements in
            ``base_raster_path_band_const_list``.

    Raises:
        ``ValueError`` if numpy array types in
            ``base_raster_path_band_const_list`` do not have sizes which can
            be broadcast against each other.
        ``ValueError`` if calculated broadcast size is incompatable with the
            raster sizes in ``base_raster_path_band_const_list``.
        ``ValueError`` if only ``'raw'`` objects have been passed as arguments.

    """
    numpy_broadcast_list = [
        x for x in base_raster_path_band_const_list
        if isinstance(x, numpy.ndarray)]
    numpy_broadcast_size = None

    try:
        # numpy.broadcast can only take up to 32 arguments, this loop works
        # around that restriction:
        while len(numpy_broadcast_list) > 1:
            numpy_broadcast_list = (
                [numpy.broadcast(*numpy_broadcast_list[:32])] +
                numpy_broadcast_list[32:])
        if numpy_broadcast_list:
            numpy_broadcast_size = numpy_broadcast_list[0].shape
    except ValueError:
        # this gets raised if numpy.broadcast fails
        raise ValueError(
            "Numpy array inputs cannot be broadcast into a single shape %s" %
            numpy_broadcast_list)

    if numpy_broadcast_list and len(numpy_broadcast_list[0].shape) > 2:
        raise ValueError(
            "Numpy array inputs must be 2 dimensions or less %s" %
            numpy_broadcast_list)

    # if there are both rasters and arrays, check the numpy shape will
    # be broadcastable with raster shape
    if raster_info_list and numpy_broadcast_size:
        # geospatial lists x/y order and numpy does y/x so reverse size list
        raster_shape = tuple(reversed(raster_info_list[0]['raster_size']))
        invalid_broadcast_size = False
        if len(numpy_broadcast_size) == 1:
            # if there's only one dimension it should match the last
            # dimension first, in the raster case this is the columns
            # because of the row/column order of numpy. No problem if
            # that value is ``1`` because it will be broadcast, otherwise
            # it should be the same as the raster.
            if (numpy_broadcast_size[0] != raster_shape[1] and
                    numpy_broadcast_size[0] != 1):
                invalid_broadcast_size = True
        else:
            for dim_index in range(2):
                # no problem if 1 because it'll broadcast, otherwise must
                # be the same value
                if (numpy_broadcast_size[dim_index] !=
                        raster_shape[dim_index] and
                        numpy_broadcast_size[dim_index] != 1):
                    invalid_broadcast_size = True
        if invalid_broadcast_size:
            raise ValueError(
                "Raster size %s cannot be broadcast to numpy shape %s" % (
                    raster_shape, numpy_broadcast_size))

    # create target raster
    if raster_info_list:
        # if rasters are passed, the target is the same size as the raster
        n_cols, n_rows = raster_info_list[0]['raster_size']
    elif numpy_broadcast_size:
        # numpy arrays in args and no raster result is broadcast shape
        # expanded to two dimensions if necessary
        if len(numpy_broadcast_size) == 1:
            n_rows, n_cols = 1, numpy_broadcast_size[0]
        else:
            n_rows, n_cols = numpy_broadcast_size
    else:
        raise ValueError(
            "Only (object, 'raw') values have been passed. Raster "
            "calculator requires at least a raster or numpy array as a "
            "parameter. This is the input list: %s" % pprint.pformat(
                base_raster_path_band_const_list))

    return n_cols, n_rows


def _validate_raster_input(
        base_raster_path_band_const_list, raster_info_list,
        target_raster_path):
    """Check for valid raster/arg inputs and output.

    Args:
        base_raster_path_band_const_list (list/tuple): the same object passed
            to .raster_calculator indicating the datastack to process.
        target_raster_path (str): desired target raster path from
            raster_calculator, used to ensure it is not also an input parameter

    Returns:
        None

    Raises:
        ValueError if any input parameter would cause an error when passing to
            .raster_calculator
    """
    if not base_raster_path_band_const_list:
        raise ValueError(
            "`base_raster_path_band_const_list` is empty and "
            "should have at least one value.")

    # It's a common error to not pass in path/band tuples, so check for that
    # and report error if so
    bad_raster_path_list = False
    if not isinstance(base_raster_path_band_const_list, (list, tuple)):
        bad_raster_path_list = True
    else:
        for value in base_raster_path_band_const_list:
            if (not _is_raster_path_band_formatted(value) and
                not isinstance(value, numpy.ndarray) and
                not (isinstance(value, tuple) and len(value) == 2 and
                     value[1] == 'raw')):
                bad_raster_path_list = True
                break
    if bad_raster_path_list:
        raise ValueError(
            "Expected a sequence of path / integer band tuples, "
            "ndarrays, or (value, 'raw') pairs for "
            "`base_raster_path_band_const_list`, instead got: "
            "%s" % pprint.pformat(base_raster_path_band_const_list))

    with GDALUseExceptions():
        # check that any rasters exist on disk and have enough bands
        not_found_paths = []
        gdal.PushErrorHandler('CPLQuietErrorHandler')
        base_raster_path_band_list = [
            path_band for path_band in base_raster_path_band_const_list
            if _is_raster_path_band_formatted(path_band)]
        for value in base_raster_path_band_list:
            if gdal.OpenEx(value[0], gdal.OF_RASTER) is None:
                not_found_paths.append(value[0])
        gdal.PopErrorHandler()
        if not_found_paths:
            raise ValueError(
                "The following files were expected but do not exist on the "
                "filesystem: " + str(not_found_paths))

        # check that band index exists in raster
        invalid_band_index_list = []
        for value in base_raster_path_band_list:
            raster = gdal.OpenEx(value[0], gdal.OF_RASTER)
            if not (1 <= value[1] <= raster.RasterCount):
                invalid_band_index_list.append(value)
            raster = None
        if invalid_band_index_list:
            raise ValueError(
                "The following rasters do not contain requested band "
                "indexes: %s" % invalid_band_index_list)

    # check that the target raster is not also an input raster
    if target_raster_path in [x[0] for x in base_raster_path_band_list]:
        raise ValueError(
            "%s is used as a target path, but it is also in the base input "
            "path list %s" % (
                target_raster_path, str(base_raster_path_band_const_list)))

    # check that raster inputs are all the same dimensions
    geospatial_info_set = set()
    for raster_info in raster_info_list:
        geospatial_info_set.add(raster_info['raster_size'])
    if len(geospatial_info_set) > 1:
        raise ValueError(
            "Input Rasters are not the same dimensions. The "
            "following raster are not identical %s" % str(
                geospatial_info_set))


def raster_calculator(
        base_raster_path_band_const_list, local_op, target_raster_path,
        datatype_target, nodata_target,
        n_workers=max(1, multiprocessing.cpu_count()),
        calc_raster_stats=True, use_shared_memory=False,
        largest_block=_LARGEST_ITERBLOCK,
        raster_driver_creation_tuple=DEFAULT_GTIFF_CREATION_TUPLE_OPTIONS):
    """Apply local a raster operation on a stack of rasters.

    This function applies a user defined function across a stack of
    rasters' pixel stack. The rasters in ``base_raster_path_band_list`` must
    be spatially aligned and have the same cell sizes.

    Args:
        base_raster_path_band_const_list (sequence): a sequence containing
            either (str, int) tuples, ``numpy.ndarray`` s of up to two
            dimensions, or an (object, 'raw') tuple.  A ``(str, int)``
            tuple refers to a raster path band index pair to use as an input.
            The ``numpy.ndarray`` s must be broadcastable to each other AND the
            size of the raster inputs. Values passed by  ``(object, 'raw')``
            tuples pass ``object`` directly into the ``local_op``. All rasters
            must have the same raster size. If only arrays are input, numpy
            arrays must be broadcastable to each other and the final raster
            size will be the final broadcast array shape. A value error is
            raised if only "raw" inputs are passed. Paths may use any
            GDAL-supported scheme, including virtual file system /vsi schemes.
        local_op (function) a function that must take in as many parameters as
            there are elements in ``base_raster_path_band_const_list``. The
            parameters in ``local_op`` will map 1-to-1 in order with the values
            in ``base_raster_path_band_const_list``. ``raster_calculator`` will
            call ``local_op`` to generate the pixel values in ``target_raster``
            along memory block aligned processing windows. Note any
            particular call to ``local_op`` will have the arguments from
            ``raster_path_band_const_list`` sliced to overlap that window.
            If an argument from ``raster_path_band_const_list`` is a
            raster/path band tuple, it will be passed to ``local_op`` as a 2D
            numpy array of pixel values that align with the processing window
            that ``local_op`` is targeting. A 2D or 1D array will be sliced to
            match the processing window and in the case of a 1D array tiled in
            whatever dimension is flat. If an argument is a scalar it is
            passed as as scalar.
            The return value must be a 2D array of the same size as any of the
            input parameter 2D arrays and contain the desired pixel values
            for the target raster.
        target_raster_path (string): the path of the output raster.  The
            projection, size, and cell size will be the same as the rasters
            in ``base_raster_path_const_band_list`` or the final broadcast
            size of the constant/ndarray values in the list.
        datatype_target (gdal datatype; int): the desired GDAL output type of
            the target raster.
        nodata_target (numerical value): the desired nodata value of the
            target raster.
        n_workers (int): number of Processes to launch for parallel processing,
            defaults to ``multiprocessing.cpu_count()``.
        calc_raster_stats (boolean): If True, calculates and sets raster
            statistics (min, max, mean, and stdev) for target raster.
        use_shared_memory (boolean): If True, uses Python Multiprocessing
            shared memory to calculate raster stats for faster performance.
            This feature is available for Python >= 3.8 and will otherwise
            be ignored for earlier versions of Python.
        largest_block (int): Attempts to internally iterate over raster blocks
            with this many elements.  Useful in cases where the blocksize is
            relatively small, memory is available, and the function call
            overhead dominates the iteration.  Defaults to 2**20.  A value of
            anything less than the original blocksize of the raster will
            result in blocksizes equal to the original size.
        raster_driver_creation_tuple (tuple): a tuple containing a GDAL driver
            name string as the first element and a GDAL creation options
            tuple/list as the second. Defaults to
            geoprocessing.DEFAULT_GTIFF_CREATION_TUPLE_OPTIONS.

    Returns:
        None

    Raises:
        ValueError: invalid input provided

    """
    raster_info_list = [
        get_raster_info(path_band[0])
        for path_band in base_raster_path_band_const_list
        if _is_raster_path_band_formatted(path_band)]

    target_raster_path = os.path.abspath(target_raster_path)
    _validate_raster_input(
        base_raster_path_band_const_list, raster_info_list, target_raster_path)

    n_cols, n_rows = _calculate_target_raster_size(
        raster_info_list, base_raster_path_band_const_list)

    # create a "canonical" argument list that contains only
    # (file paths, band id) tuples, 2d numpy arrays, or raw values
    base_canonical_arg_list = []
    for value in base_raster_path_band_const_list:
        # the input has been tested and value is either a raster/path band
        # tuple, 1d ndarray, 2d ndarray, or (value, 'raw') tuple.
        if _is_raster_path_band_formatted(value):
            # it's a raster/path band, keep track of open raster and band
            # for later so we can `None` them.
            base_canonical_arg_list.append(
                RasterPathBand(value[0], value[1]))
        elif isinstance(value, numpy.ndarray):
            if value.ndim == 1:
                # easier to process as a 2d array for writing to band
                base_canonical_arg_list.append(
                    value.reshape((1, value.shape[0])))
            else:  # dimensions are two because we checked earlier.
                base_canonical_arg_list.append(value)
        else:
            # it's a regular tuple
            base_canonical_arg_list.append(value)

    if datatype_target not in _VALID_GDAL_TYPES:
        raise ValueError(
            'Invalid target type, should be a gdal.GDT_* type, received '
            '"%s"' % datatype_target)

    with GDALUseExceptions():
        # create target raster
        raster_driver = gdal.GetDriverByName(raster_driver_creation_tuple[0])
        try:
            os.makedirs(os.path.dirname(target_raster_path))
        except OSError as exception:
            # it's fine if the directory already exists, otherwise there's a big
            # error!
            if exception.errno != errno.EEXIST:
                raise

        target_raster = raster_driver.Create(
            target_raster_path, n_cols, n_rows, 1, datatype_target,
            options=raster_driver_creation_tuple[1])

        target_band = target_raster.GetRasterBand(1)
        if nodata_target is not None:
            target_band.SetNoDataValue(nodata_target)
        if raster_info_list:
            # use the first raster in the list for the projection and geotransform
            target_raster.SetProjection(raster_info_list[0]['projection_wkt'])
            target_raster.SetGeoTransform(raster_info_list[0]['geotransform'])
        target_band = None
        target_raster = None

    manager = multiprocessing.Manager()
    stats_worker_queue = None
    if calc_raster_stats:
        # if this queue is used to send computed valid blocks of
        # the raster to an incremental statistics calculator worker
        stats_worker_queue = manager.Queue()

    # iterate over each block and calculate local_op
    block_offset_list = list(iterblocks(
        (target_raster_path, 1), offset_only=True,
        largest_block=largest_block))

    if calc_raster_stats:
        LOGGER.debug('start stats worker')
        stats_worker = multiprocessing.Process(
            target=geoprocessing_core.stats_worker,
            args=(stats_worker_queue, len(block_offset_list)))
        stats_worker.start()

    LOGGER.debug('start workers')
    processing_state = manager.dict()
    processing_state['blocks_complete'] = 0
    processing_state['total_blocks'] = len(block_offset_list)
    processing_state['last_time'] = time.time()
    block_size_bytes = (
        numpy.dtype(numpy.float64).itemsize *
        block_offset_list[0]['win_xsize'] * block_offset_list[0]['win_ysize'])
    target_write_lock = manager.Lock()
    block_offset_queue = multiprocessing.Queue(n_workers)
    process_list = []
    for _ in range(n_workers):
        shared_memory = None
        if calc_raster_stats:
            if sys.version_info >= (3, 8) and use_shared_memory:
                shared_memory = multiprocessing.shared_memory.SharedMemory(
                    create=True, size=block_size_bytes)
        worker = multiprocessing.Process(
            target=_raster_calculator_worker,
            args=(
                block_offset_queue, base_canonical_arg_list, local_op,
                stats_worker_queue, nodata_target, target_raster_path,
                target_write_lock, processing_state, shared_memory))
        worker.start()
        process_list.append((worker, shared_memory))

    # Fill the work queue
    for block_offset in block_offset_list:
        block_offset_queue.put(block_offset)

    for _ in range(n_workers):
        block_offset_queue.put(None)

    LOGGER.info('wait for stats worker to complete')
    stats_worker.join(_MAX_TIMEOUT)
    if stats_worker.is_alive():
        LOGGER.error(
            f'stats worker {stats_worker.pid} '
            'didn\'t terminate, sending kill signal.')
        try:
            os.kill(stats_worker.pid, signal.SIGTERM)
        except Exception:
            LOGGER.exception(f'unable to kill {stats_worker.pid}')

    # wait for the workers to join
    LOGGER.info('all work sent, waiting for workers to finish')
    for worker, shared_memory in process_list:
        worker.join(_MAX_TIMEOUT)
        if worker.is_alive():
            LOGGER.error(
                f'worker {worker.pid} didn\'t terminate, sending kill signal.')
            try:
                os.kill(stats_worker.pid, signal.SIGTERM)
            except Exception:
                LOGGER.exception(f'unable to kill {worker.pid}')
        if shared_memory is not None:
            LOGGER.debug(f'unlink {shared_memory.name}')
            shared_memory.unlink()

    if calc_raster_stats:
        payload = stats_worker_queue.get(True, _MAX_TIMEOUT)
        if payload is not None:
            target_min, target_max, target_mean, target_stddev = payload
            with GDALUseExceptions():
                target_raster = gdal.OpenEx(
                    target_raster_path, gdal.OF_RASTER | gdal.GA_Update)
                target_band = target_raster.GetRasterBand(1)
                target_band.SetStatistics(
                    float(target_min), float(target_max), float(target_mean),
                    float(target_stddev))
                target_band = None
                target_raster = None
    LOGGER.info('raster_calculator 100.0%% complete')



import logging
import os
import shutil
import tempfile
import time

cimport cython
cimport numpy
numpy.import_array()
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from cython.operator cimport dereference as deref
from cython.operator cimport preincrement as inc
from libc.time cimport time as ctime
from libc.time cimport time_t
from libcpp.list cimport list as clist
from libcpp.map cimport map as cmap
from libcpp.pair cimport pair
from libcpp.queue cimport queue
from libcpp.set cimport set as cset
from osgeo import gdal
from osgeo import ogr
from osgeo import osr
import numpy
import shapely.geometry
import shapely.prepared
import shapely.wkb

import pygeoprocessing
from ..geoprocessing_core import SPARSE_CREATION_OPTIONS
from ..extensions cimport ManagedRaster
from ..utils import gdal_use_exceptions, GDALUseExceptions

LOGGER = logging.getLogger(__name__)

# It's convenient to define a C++ pair here as a pair of longs to represent the
# x,y coordinates of a pixel.  So, CoordinatePair().first is the x coordinate,
# CoordinatePair().second is the y coordinate.  Both are in integer pixel
# coordinates.
ctypedef pair[long, long] CoordinatePair


def _is_raster_path_band_formatted(raster_path_band):
    """Return true if raster path band is a (str, int) tuple/list."""
    if not isinstance(raster_path_band, (list, tuple)):
        return False
    elif len(raster_path_band) != 2:
        return False
    elif not isinstance(raster_path_band[0], basestring):
        return False
    elif not isinstance(raster_path_band[1], int):
        return False
    else:
        return True


cdef cset[CoordinatePair] _c_split_geometry_into_seeds(
        source_geom_wkb, tuple flow_dir_geotransform, flow_dir_srs,
        flow_dir_n_cols, flow_dir_n_rows,
        target_raster_path, diagnostic_vector_path=None):
    """Split a geometry into 'seeds' of (x, y) coordinate pairs.

    Parameters:
        source_geom_wkb (str): A string of bytes in WKB representing
            a geometry. Must be in the same projected coordinate system
            as the flow direction raster from which
            ``flow_dir_geotransform`` and ``flow_dir_srs`` are derived.
        flow_dir_geotransform (list): A 6-element array representing
            the GDAL affine geotransform from the flow direction raster.
        flow_dir_srs (osr.SpatialReference): The OSR SpatialReference
            object from the flow direction raster.
        flow_dir_n_cols (int): the number of columns in the flow
            direction raster.
        flow_dir_n_rows (int): the number of rows in the flow
            direction raster.
        target_raster_path (str): The path to a raster onto which
            the geometry might be rasterized.  If the geometry is small
            enough to be completely contained within a single pixel, no
            raster will be written to this location.
        diagnostic_vector_path (string or None): If a string, a GeoPackage
            vector will be written to this path containing georeferenced
            points representing the 'seeds' determined by this function.
            If ``None``, no vector is created.

    Returns:
        A ``cset[CoordinatePair]`` for use in cython code.

    """
    cdef float minx, miny, maxx, maxy
    cdef double x_origin = flow_dir_geotransform[0]
    cdef double y_origin = flow_dir_geotransform[3]
    cdef double x_pixelwidth = flow_dir_geotransform[1]
    cdef double y_pixelwidth = flow_dir_geotransform[5]
    cdef double flow_dir_maxx = x_origin + (flow_dir_n_cols * x_pixelwidth)
    cdef double flow_dir_miny = y_origin + (flow_dir_n_rows * y_pixelwidth)
    cdef cset[CoordinatePair] seed_set

    geometry = shapely.wkb.loads(source_geom_wkb)

    minx, miny, maxx, maxy = geometry.bounds
    cdef int minx_pixelcoord = <int>((minx - x_origin) // x_pixelwidth)
    cdef int miny_pixelcoord = <int>((miny - y_origin) // y_pixelwidth)
    cdef int maxx_pixelcoord = <int>((maxx - x_origin) // x_pixelwidth)
    cdef int maxy_pixelcoord = <int>((maxy - y_origin) // y_pixelwidth)

    # If the geometry only intersects a single pixel, we can treat it
    # as a single point, which means that we can track it directly in our
    # seeds data structure and not have to include it in the disjoint set
    # determination.
    if minx_pixelcoord == maxx_pixelcoord and miny_pixelcoord == maxy_pixelcoord:
        # If the point is over nodata, skip it.
        seed = CoordinatePair(minx_pixelcoord, miny_pixelcoord)
        seed_set.insert(seed)
        return seed_set

    # If the geometry's bounding box covers more than one pixel, we need to
    # rasterize it to determine which pixels it intersects.
    cdef double minx_aligned = max(
        x_origin + (minx_pixelcoord * x_pixelwidth),
        x_origin)
    cdef double miny_aligned = max(
        y_origin + ((miny_pixelcoord+1) * y_pixelwidth),
        flow_dir_miny)
    cdef double maxx_aligned = min(
        x_origin + ((maxx_pixelcoord+1) * x_pixelwidth),
        flow_dir_maxx)
    cdef double maxy_aligned = min(
        y_origin + (maxy_pixelcoord * y_pixelwidth),
        y_origin)
    cdef int write_diagnostic_vector = 0

    cdef int row, col
    cdef int global_row, global_col
    cdef int seed_raster_origin_col
    cdef int seed_raster_origin_row
    cdef dict block_info
    cdef int block_xoff
    cdef int block_yoff
    cdef numpy.ndarray[numpy.npy_uint8, ndim=2] seed_array

    # It's possible for a perfectly vertical or horizontal line to cover 0 rows
    # or columns, so defaulting to row/col count of 1 in these cases.
    local_n_cols = max(abs(maxx_aligned - minx_aligned) // abs(x_pixelwidth), 1)
    local_n_rows = max(abs(maxy_aligned - miny_aligned) // abs(y_pixelwidth), 1)

    with GDALUseExceptions():
        # The geometry does not fit into a single pixel, so let's create a new
        # raster onto which to rasterize it.
        memory_driver = gdal.GetDriverByName('Memory')
        new_vector = memory_driver.Create('mem', 0, 0, 0, gdal.GDT_Unknown)
        new_layer = new_vector.CreateLayer('user_geometry', flow_dir_srs, ogr.wkbUnknown)
        new_layer.StartTransaction()
        new_feature = ogr.Feature(new_layer.GetLayerDefn())
        new_feature.SetGeometry(ogr.CreateGeometryFromWkb(source_geom_wkb))
        new_layer.CreateFeature(new_feature)
        new_layer.CommitTransaction()

        local_origin_x = max(minx_aligned, x_origin)
        local_origin_y = min(maxy_aligned, y_origin)

        local_geotransform = [
            local_origin_x, flow_dir_geotransform[1], flow_dir_geotransform[2],
            local_origin_y, flow_dir_geotransform[4], flow_dir_geotransform[5]]
        gtiff_driver = gdal.GetDriverByName('GTiff')
        # Raster is sparse, no need to fill.
        raster = gtiff_driver.Create(
            target_raster_path, int(local_n_cols), int(local_n_rows), 1,
            gdal.GDT_Byte, options=SPARSE_CREATION_OPTIONS)

        raster.SetSpatialRef(flow_dir_srs)
        raster.SetGeoTransform(local_geotransform)

        gdal.RasterizeLayer(
            raster, [1], new_layer, burn_values=[1], options=['ALL_TOUCHED=True'])
        raster = None

        diagnostic_vector = None
        diagnostic_layer = None
        if diagnostic_vector_path is not None:
            write_diagnostic_vector = 1

            gpkg_driver = gdal.GetDriverByName('GPKG')
            diagnostic_vector = gpkg_driver.Create(
                diagnostic_vector_path, 0, 0, 0, gdal.GDT_Unknown)
            diagnostic_layer = diagnostic_vector.CreateLayer(
                'seeds', flow_dir_srs, ogr.wkbPoint)
            user_geometry_layer = diagnostic_vector.CreateLayer(
                'user_geometry', flow_dir_srs, ogr.wkbUnknown)
            user_geometry_layer.StartTransaction()
            user_feature = ogr.Feature(user_geometry_layer.GetLayerDefn())
            user_feature.SetGeometry(ogr.CreateGeometryFromWkb(source_geom_wkb))
            user_geometry_layer.CreateFeature(user_feature)
            user_geometry_layer.CommitTransaction()

        seed_raster_origin_col = <int>((local_origin_x - x_origin) // x_pixelwidth)
        seed_raster_origin_row = <int>((local_origin_y - y_origin) // y_pixelwidth)
        for block_info, seed_array in pygeoprocessing.iterblocks(
                (target_raster_path, 1)):
            block_xoff = block_info['xoff']
            block_yoff = block_info['yoff']
            n_rows = seed_array.shape[0]
            n_cols = seed_array.shape[1]

            for row in range(n_rows):
                for col in range(n_cols):
                    with cython.boundscheck(False):
                        # Check if the pixel does not overlap the geometry.
                        if seed_array[row, col] == 0:
                            continue

                    global_row = seed_raster_origin_row + block_yoff + row
                    global_col = seed_raster_origin_col + block_xoff + col

                    if write_diagnostic_vector == 1:
                        diagnostic_layer.StartTransaction()
                        new_feature = ogr.Feature(diagnostic_layer.GetLayerDefn())
                        new_feature.SetGeometry(ogr.CreateGeometryFromWkb(
                            shapely.geometry.Point(
                                x_origin + ((global_col*x_pixelwidth) +
                                            (x_pixelwidth / 2.)),
                                y_origin + ((global_row*y_pixelwidth) +
                                            (y_pixelwidth / 2.))).wkb))
                        diagnostic_layer.CreateFeature(new_feature)
                        diagnostic_layer.CommitTransaction()

                    seed_set.insert(CoordinatePair(global_col, global_row))

        return seed_set


def _split_geometry_into_seeds(
        source_geom_wkb, geotransform, flow_dir_srs,
        flow_dir_n_cols, flow_dir_n_rows, target_raster_path,
        diagnostic_vector_path=None):
    """Split a geometry into 'seeds' of (x, y) coordinate pairs.

    This function is a python wrapper around ``_c_split_geometry_into_seeds``
    that is useful for testing.

    Parameters:
        source_geom_wkb (str): A string of bytes in WKB representing
            a geometry. Must be in the same projected coordinate system
            as the flow direction raster from which
            ``flow_dir_geotransform`` and ``flow_dir_srs`` are derived.
        flow_dir_geotransform (list): A 6-element array representing
            the GDAL affine geotransform from the flow direction raster.
        flow_dir_srs (osr.SpatialReference): The OSR SpatialReference
            object from the flow direction raster.
        flow_dir_n_cols (int): the number of columns in the flow
            direction raster.
        flow_dir_n_rows (int): the number of rows in the flow
            direction raster.
        target_raster_path (str): The path to a raster onto which
            the geometry might be rasterized.  If the geometry is small
            enough to be completely contained within a single pixel, no
            raster will be written to this location.
        diagnostic_vector_path (string or None): If a string, a GeoPackage
            vector will be written to this path containing georeferenced
            points representing the 'seeds' determined by this function.
            If ``None``, no vector is created.

    Returns:
        A python ``set`` of (x-index, y-index) tuples.

    """

    return_set = set()
    cdef cset[CoordinatePair] seeds = _c_split_geometry_into_seeds(
        source_geom_wkb, geotransform, flow_dir_srs, flow_dir_n_cols,
        flow_dir_n_rows, target_raster_path, diagnostic_vector_path)

    cdef CoordinatePair seed
    cdef cset[CoordinatePair].iterator seeds_iterator = seeds.begin()
    while seeds_iterator != seeds.end():
        seed = deref(seeds_iterator)
        inc(seeds_iterator)

        return_set.add((seed.first, seed.second))

    return return_set


@cython.boundscheck(False)
@gdal_use_exceptions
def delineate_watersheds_d8(
        d8_flow_dir_raster_path_band, outflow_vector_path,
        target_watersheds_vector_path, working_dir=None,
        write_diagnostic_vector=False, remove_temp_files=True,
        target_layer_name='watersheds'):
    """Delineate watersheds for a vector of geometries using D8 flow dir.

    Note:
        The ``d8_flow_dir_raster_path_band`` and ``outflow_vector_path`` files
        must have the same spatial reference system.  The output watersheds
        vector will use this same spatial reference system.

    Args:
        d8_flow_dir_raster_path_band (tuple): A (path, band_id) tuple
            to a D8 flow direction raster.  This raster must be a tiled raster
            with block sizes being a power of 2.  The output watersheds vector
            will have its spatial reference copied from this raster. Paths may
            use any GDAL-supported scheme, including virtual file system /vsi schemes.
        outflow_vector_path (str): The path to a vector containing features
            with valid geometries from which watersheds will be delineated.
            Only those parts of the geometry that overlap valid flow direction
            pixels will be included in the output watersheds vector. Paths may
            use any GDAL-supported scheme, including virtual file system /vsi schemes.
        target_watersheds_vector_path (str): The path to a vector where the
            target watersheds will be stored.  Must have the extension ``.gpkg``.
        working_dir=None (str or None): The path to a directory on disk
            within which various intermediate files will be stored.  If None,
            a folder will be created within the system's temp directory.
        write_diagnostic_vector=False (bool): If ``True``, a set of vectors will
            be written to ``working_dir``, one per watershed.  Each vector
            includes geometries for the watershed being represented and
            for the watershed seed pixels the geometry overlaps.  Useful in
            debugging issues with feature overlap of the DEM.  Setting this
            parameter to ``True`` will dramatically increase runtime when
            outflow geometries cover many pixels.
        remove_temp_files=True (bool): Whether to remove the created temp
            directory at the end of the watershed delineation run.
        target_layer_name='watersheds' (str): The string name to use for
            the watersheds layer.  This layer name may be named anything
            except for "polygonized_watersheds".

    Returns:
        ``None``

    """
    try:
        if working_dir is not None:
            os.makedirs(working_dir)
    except OSError:
        pass
    working_dir_path = tempfile.mkdtemp(
        dir=working_dir, prefix='watershed_delineation_trivial_%s_' % time.strftime(
            '%Y-%m-%d_%H_%M_%S', time.gmtime()))

    if (d8_flow_dir_raster_path_band is not None and not
            _is_raster_path_band_formatted(d8_flow_dir_raster_path_band)):
        raise ValueError(
            "%s is supposed to be a raster band tuple but it's not." % (
                d8_flow_dir_raster_path_band))

    flow_dir_info = pygeoprocessing.get_raster_info(
        d8_flow_dir_raster_path_band[0])
    cdef int flow_dir_nodata = flow_dir_info['nodata'][0]
    source_gt = flow_dir_info['geotransform']
    cdef double flow_dir_origin_x = source_gt[0]
    cdef double flow_dir_origin_y = source_gt[3]
    cdef double flow_dir_pixelsize_x = source_gt[1]
    cdef double flow_dir_pixelsize_y = source_gt[5]
    cdef int flow_dir_n_cols = flow_dir_info['raster_size'][0]
    cdef int flow_dir_n_rows = flow_dir_info['raster_size'][1]
    cdef int ws_id
    bbox_minx, bbox_miny, bbox_maxx, bbox_maxy = flow_dir_info['bounding_box']
    LOGGER.debug('Creating flow dir bbox')
    flow_dir_bbox = shapely.prepared.prep(
        shapely.geometry.Polygon([
            (bbox_minx, bbox_maxy),
            (bbox_minx, bbox_miny),
            (bbox_maxx, bbox_miny),
            (bbox_maxx, bbox_maxy),
            (bbox_minx, bbox_maxy)]))
    LOGGER.debug('Creating flow dir managed raster')
    flow_dir_managed_raster = ManagedRaster(
        d8_flow_dir_raster_path_band[0].encode('utf-8'),
        d8_flow_dir_raster_path_band[1], False)
    gtiff_driver = gdal.GetDriverByName('GTiff')
    flow_dir_srs = osr.SpatialReference()
    if flow_dir_info['projection_wkt']:
        flow_dir_srs.ImportFromWkt(flow_dir_info['projection_wkt'])

    outflow_vector = gdal.OpenEx(outflow_vector_path, gdal.OF_VECTOR)
    if outflow_vector is None:
        raise ValueError(u'Could not open outflow vector %s' % outflow_vector_path)

    driver = ogr.GetDriverByName('GPKG')
    watersheds_srs = osr.SpatialReference()
    if flow_dir_info['projection_wkt']:
        watersheds_srs.ImportFromWkt(flow_dir_info['projection_wkt'])
    watersheds_vector = driver.CreateDataSource(target_watersheds_vector_path)
    polygonized_watersheds_layer = watersheds_vector.CreateLayer(
        'polygonized_watersheds', watersheds_srs, ogr.wkbPolygon)

    # Using wkbMultiPolygon for this layer because GDAL's polygonize function
    # may create multiple polygons for a single watershed feature, and we want
    # all geometries for a single watershed to be represented in a single
    # feature.
    watersheds_layer = watersheds_vector.CreateLayer(
        target_layer_name, watersheds_srs, ogr.wkbMultiPolygon)
    index_field = ogr.FieldDefn('ws_id', ogr.OFTInteger)
    index_field.SetWidth(24)
    polygonized_watersheds_layer.CreateField(index_field)

    cdef int* reverse_flow = [4, 5, 6, 7, 0, 1, 2, 3]
    cdef int* neighbor_col = [1, 1, 0, -1, -1, -1, 0, 1]
    cdef int* neighbor_row = [0, -1, -1, -1, 0, 1, 1, 1]
    cdef queue[CoordinatePair] process_queue
    cdef cset[CoordinatePair] process_queue_set
    cdef CoordinatePair neighbor_pixel
    cdef int ix_min, iy_min, ix_max, iy_max
    cdef ManagedRaster scratch_managed_raster
    cdef int watersheds_created = 0
    cdef int current_fid, outflow_feature_count
    cdef cset[CoordinatePair].iterator seed_iterator
    cdef cset[CoordinatePair] seeds_in_watershed
    cdef time_t last_log_time
    cdef int n_cells_visited = 0

    LOGGER.info('Delineating watersheds')
    outflow_layer = outflow_vector.GetLayer()
    outflow_feature_count = outflow_layer.GetFeatureCount()
    for feature in outflow_layer:
        # Some vectors start indexing their FIDs at 0.
        # The mask raster input to polygonization, however, only regards pixels
        # as zero or nonzero.  Therefore, to make sure we can use the ws_id as
        # the FID and not maintain a separate mask raster, we'll just add 1.
        current_fid = feature.GetFID()
        ws_id = current_fid + 1
        assert ws_id >= 1, 'WSID <= 1!'

        geom = feature.GetGeometryRef()
        if geom.IsEmpty():
            LOGGER.debug(
                'Outflow feature %s has empty geometry.  Skipping.',
                current_fid)
            continue
        geom_wkb = bytes(geom.ExportToWkb())
        shapely_geom = shapely.wkb.loads(geom_wkb)

        LOGGER.debug('Testing geometry bbox')
        if not flow_dir_bbox.intersects(shapely.geometry.box(*shapely_geom.bounds)):
            LOGGER.debug(
                'Outflow feature %s does not overlap with the flow '
                'direction raster. Skipping.', current_fid)
            continue

        seeds_raster_path = os.path.join(working_dir_path, '%s_rasterized.tif' % ws_id)
        if write_diagnostic_vector:
            diagnostic_vector_path = os.path.join(working_dir_path, '%s_seeds.gpkg' % ws_id)
        else:
            diagnostic_vector_path = None
        seeds_in_watershed = _c_split_geometry_into_seeds(
            geom_wkb,
            source_gt,
            flow_dir_srs=flow_dir_srs,
            flow_dir_n_cols=flow_dir_n_cols,
            flow_dir_n_rows=flow_dir_n_rows,
            target_raster_path=seeds_raster_path,
            diagnostic_vector_path=diagnostic_vector_path
        )

        seed_iterator = seeds_in_watershed.begin()
        while seed_iterator != seeds_in_watershed.end():
            seed = deref(seed_iterator)
            inc(seed_iterator)

            if not 0 <= seed.first < flow_dir_n_cols:
                continue

            if not 0 <= seed.second < flow_dir_n_rows:
                continue

            if flow_dir_managed_raster.get(seed.first, seed.second) == flow_dir_nodata:
                continue

            process_queue.push(seed)
            process_queue_set.insert(seed)

        if process_queue_set.size() == 0:
            LOGGER.debug(
                'Outflow feature %s does not intersect any pixels with '
                'valid flow direction. Skipping.', current_fid)
            continue

        scratch_raster_path = os.path.join(working_dir_path,
                                           '%s_scratch.tif' % ws_id)
        scratch_raster = gtiff_driver.Create(
            scratch_raster_path,
            flow_dir_n_cols,
            flow_dir_n_rows,
            1,  # n bands
            gdal.GDT_UInt32,
            options=SPARSE_CREATION_OPTIONS)
        scratch_raster.SetGeoTransform(source_gt)
        if flow_dir_info['projection_wkt']:
            scratch_raster.SetProjection(flow_dir_info['projection_wkt'])
        # strictly speaking, there's no need to set the nodata value on the band.
        scratch_raster = None

        scratch_managed_raster = ManagedRaster(
            scratch_raster_path.encode('utf-8'), 1, True)
        ix_min = flow_dir_n_cols
        iy_min = flow_dir_n_rows
        ix_max = 0
        iy_max = 0
        n_cells_visited = 0
        LOGGER.info(
            'Delineating watershed %s of %s (ws_id %s)',
            current_fid, outflow_feature_count, ws_id)
        last_log_time = ctime(NULL)
        while not process_queue.empty():
            if ctime(NULL) - last_log_time > 5.0:
                last_log_time = ctime(NULL)
                LOGGER.info(
                    'Delineating watershed %s of %s (ws_id %s), %s pixels '
                    'found so far', current_fid, outflow_feature_count,
                    ws_id, n_cells_visited)

            current_pixel = process_queue.front()
            process_queue_set.erase(current_pixel)
            process_queue.pop()

            scratch_managed_raster.set(current_pixel.first,
                                       current_pixel.second, ws_id)
            n_cells_visited += 1

            # These are for tracking the extents of the raster so we can build
            # a VRT and only polygonize the pixels we need to.
            ix_min = min(ix_min, current_pixel.first)
            iy_min = min(iy_min, current_pixel.second)
            ix_max = max(ix_max, current_pixel.first)
            iy_max = max(iy_max, current_pixel.second)

            for neighbor_index in range(8):
                neighbor_pixel = CoordinatePair(
                    current_pixel.first + neighbor_col[neighbor_index],
                    current_pixel.second + neighbor_row[neighbor_index])

                if not 0 <= neighbor_pixel.first < flow_dir_n_cols:
                    continue

                if not 0 <= neighbor_pixel.second < flow_dir_n_rows:
                    continue

                # If we've already enqueued the neighbor (either it's
                # upstream of another pixel or it's a watershed seed), we
                # don't need to re-enqueue it.
                if (process_queue_set.find(neighbor_pixel) !=
                        process_queue_set.end()):
                    continue

                # If the neighbor is known to be a seed, we don't need to
                # re-enqueue it either.  Either it's been visited already (and
                # may have upstream pixels) or it's going to be.
                if (seeds_in_watershed.find(neighbor_pixel) !=
                        seeds_in_watershed.end()):
                    continue

                # Does the neighbor flow into this pixel?
                # If yes, enqueue it.
                if (reverse_flow[neighbor_index] ==
                        flow_dir_managed_raster.get(
                            neighbor_pixel.first, neighbor_pixel.second)):
                    process_queue.push(neighbor_pixel)
                    process_queue_set.insert(neighbor_pixel)

        watersheds_created += 1
        scratch_managed_raster.close()

        # Build a VRT from the bounds of the affected pixels before
        # rasterizing.  For watersheds significantly smaller than the flow dir
        # raster, this yields a large speedup because we don't have to read in
        # the whole scratch raster in order to polygonize.
        x1 = (flow_dir_origin_x + (max(ix_min-1, 0)*flow_dir_pixelsize_x))  # minx
        y1 = (flow_dir_origin_y + (max(iy_min-1, 0)*flow_dir_pixelsize_y))  # miny
        x2 = (flow_dir_origin_x + (min(ix_max+1, flow_dir_n_cols)*flow_dir_pixelsize_x))  # maxx
        y2 = (flow_dir_origin_y + (min(iy_max+1, flow_dir_n_rows)*flow_dir_pixelsize_y))  # maxy

        vrt_options = gdal.BuildVRTOptions(
            outputBounds=(
                min(x1, x2),
                min(y1, y2),
                max(x1, x2),
                max(y1, y2))
        )
        vrt_path = os.path.join(working_dir_path, '%s_vrt.vrt' % ws_id)
        gdal.BuildVRT(vrt_path, [scratch_raster_path], options=vrt_options)

        # Polygonize this new watershed from the VRT.
        vrt_raster = gdal.OpenEx(vrt_path, gdal.OF_RASTER, allowed_drivers=['VRT'])
        vrt_band = vrt_raster.GetRasterBand(1)
        _ = gdal.Polygonize(
            vrt_band,  # The source band
            vrt_band,  # The mask. Pixels with 0 are invalid, nonzero are valid
            polygonized_watersheds_layer,
            0,  # ws_id field index
            [])  # 8connectedness does not always produce valid geometries.
        _ = None
        vrt_band = None
        vrt_raster = None

        # Removing files as we go to help manage disk space.
        if remove_temp_files:
            os.remove(scratch_raster_path)
            if os.path.exists(seeds_raster_path):
                os.remove(seeds_raster_path)
            os.remove(vrt_path)
            if (diagnostic_vector_path
                    and os.path.exists(diagnostic_vector_path)):
                os.remove(diagnostic_vector_path)

    flow_dir_managed_raster.close()
    LOGGER.info('Finished delineating %s watersheds', watersheds_created)

    # The Polygonization algorithm will sometimes identify regions that
    # should be contiguous in a single polygon, but are not.  For this reason,
    # we need an extra consolidation step here to make sure that we only produce
    # 1 feature per watershed.
    cdef cmap[int, cset[int]] fragments_with_duplicates
    cdef int fid
    for feature in polygonized_watersheds_layer:
        fid = feature.GetFID()
        # ws_id is tracked as 1 more than the FID.  See previous note about why.
        ws_id = feature.GetField('ws_id') - 1
        if (fragments_with_duplicates.find(ws_id)
                == fragments_with_duplicates.end()):
            fragments_with_duplicates[ws_id] = cset[int]()
        fragments_with_duplicates[ws_id].insert(fid)
    polygonized_watersheds_layer.ResetReading()

    LOGGER.info(
        'Consolidating %s fragments and copying field values to '
        'watersheds layer.', polygonized_watersheds_layer.GetFeatureCount())
    source_vector = gdal.OpenEx(outflow_vector_path, gdal.OF_VECTOR)
    source_layer = source_vector.GetLayer()
    watersheds_layer.CreateFields(source_layer.schema)

    watersheds_layer.StartTransaction()
    cdef int duplicate_fid
    cdef cset[int] duplicate_ids_set
    cdef cset[int].iterator duplicate_ids_set_iterator
    cdef cmap[int, cset[int]].iterator fragments_with_duplicates_iterator
    fragments_with_duplicates_iterator = fragments_with_duplicates.begin()
    while fragments_with_duplicates_iterator != fragments_with_duplicates.end():
        ws_id = deref(fragments_with_duplicates_iterator).first
        duplicate_ids_set = deref(fragments_with_duplicates_iterator).second
        inc(fragments_with_duplicates_iterator)

        duplicate_ids_set_iterator = duplicate_ids_set.begin()

        if duplicate_ids_set.size() == 1:
            duplicate_fid = deref(duplicate_ids_set_iterator)
            source_feature = polygonized_watersheds_layer.GetFeature(duplicate_fid)
            new_geometry = source_feature.GetGeometryRef()
        else:
            new_geometry = ogr.Geometry(ogr.wkbMultiPolygon)
            while duplicate_ids_set_iterator != duplicate_ids_set.end():
                duplicate_fid = deref(duplicate_ids_set_iterator)
                inc(duplicate_ids_set_iterator)

                duplicate_feature = polygonized_watersheds_layer.GetFeature(duplicate_fid)
                duplicate_geometry = duplicate_feature.GetGeometryRef()
                new_geometry.AddGeometry(duplicate_geometry)

        watershed_feature = ogr.Feature(watersheds_layer.GetLayerDefn())
        watershed_feature.SetGeometry(ogr.ForceToMultiPolygon(new_geometry))

        source_feature = source_layer.GetFeature(ws_id)
        for field_name, field_value in source_feature.items().items():
            watershed_feature.SetField(field_name, field_value)
        watersheds_layer.CreateFeature(watershed_feature)
    watersheds_layer.CommitTransaction()

    polygonized_watersheds_layer = None
    watersheds_layer = None
    if remove_temp_files:
        watersheds_vector.DeleteLayer('polygonized_watersheds')
    LOGGER.info('Finished vector consolidation')

    watersheds_vector = None
    source_layer = None
    source_vector = None

    if remove_temp_files:
        shutil.rmtree(working_dir_path)
    LOGGER.info('Watershed delineation complete')



"""pygeoprocessing.watersheds testing suite."""
import glob
import os
import shutil
import tempfile
import unittest

import numpy
import pygeoprocessing
import pygeoprocessing.routing
import shapely.geometry
from osgeo import gdal
from osgeo import ogr
from osgeo import osr
from pygeoprocessing.routing import watershed


class WatershedDelineationTests(unittest.TestCase):
    """Main Watershed test module."""

    def setUp(self):
        """Create empty workspace dir."""
        self.workspace_dir = tempfile.mkdtemp()

    def tearDown(self):
        """Delete workspace dir."""
        shutil.rmtree(self.workspace_dir)

    def test_watersheds_diagnostic_vector(self):
        """PGP watersheds: test diagnostic vector."""
        flow_dir_array = numpy.array([
            [6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
            [6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
            [6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
            [6, 6, 6, 6, 6, 6, 6, 6, 6, 255],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 255],
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 2],
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 2],
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 2],
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 2]],
            dtype=numpy.uint8)

        srs = osr.SpatialReference()
        srs.ImportFromEPSG(32731)  # WGS84 / UTM zone 31s
        srs_wkt = srs.ExportToWkt()

        flow_dir_path = os.path.join(self.workspace_dir, 'flow_dir.tif')
        pygeoprocessing.numpy_array_to_raster(
            base_array=flow_dir_array,
            target_nodata=255,
            pixel_size=(2, -2),
            origin=(2, -2),
            projection_wkt=srs_wkt,
            target_path=flow_dir_path)

        # These geometries test:
        #  * Delineation works with varying geometry types
        #  * That we exclude seed pixels that are over nodata
        #  * That we exclude seed pixels off the bounds of the raster
        horizontal_line = shapely.geometry.LineString([(19, -11), (25, -11)])
        vertical_line = shapely.geometry.LineString([(21, -9), (21, -13)])
        square = shapely.geometry.box(17, -13, 21, -9)
        point = shapely.geometry.Point(21, -11)

        outflow_vector_path = os.path.join(self.workspace_dir, 'outflow.gpkg')
        pygeoprocessing.shapely_geometry_to_vector(
            [horizontal_line, vertical_line, square, point],
            outflow_vector_path, srs_wkt,
            'GPKG',
            {
                'polygon_id': ogr.OFTInteger,
                'field_string': ogr.OFTString,
                'other': ogr.OFTReal
            },
            [
                {'polygon_id': 1, 'field_string': 'hello world',
                 'other': 1.111},
                {'polygon_id': 2, 'field_string': 'hello foo', 'other': 2.222},
                {'polygon_id': 3, 'field_string': 'hello bar', 'other': 3.333},
                {'polygon_id': 4, 'field_string': 'hello baz', 'other': 4.444}
            ],
            ogr_geom_type=ogr.wkbUnknown)

        target_watersheds_path = os.path.join(
            self.workspace_dir, 'watersheds.gpkg')

        pygeoprocessing.routing.delineate_watersheds_d8(
            (flow_dir_path, 1), outflow_vector_path, target_watersheds_path,
            write_diagnostic_vector=True, working_dir=self.workspace_dir,
            remove_temp_files=False)

        # I'm deliberately only testing that the diagnostic files exist, not
        # the contents.  The diagnostic files should be for debugging only,
        # so I just want to make sure that they're created.
        num_diagnostic_files = len(
            glob.glob(os.path.join(self.workspace_dir, '**/*_seeds.gpkg')))
        self.assertEqual(num_diagnostic_files, 3)  # 3 features valid

    def test_watersheds_trivial(self):
        """PGP watersheds: test trivial delineation."""
        flow_dir_array = numpy.array([
            [6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
            [6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
            [6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
            [6, 6, 6, 6, 6, 6, 6, 6, 6, 255],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 255],
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 2],
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 2],
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 2],
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 2]],
            dtype=numpy.uint8)

        srs = osr.SpatialReference()
        srs.ImportFromEPSG(32731)  # WGS84 / UTM zone 31s
        srs_wkt = srs.ExportToWkt()

        for srs in (srs_wkt, None):
            flow_dir_path = os.path.join(self.workspace_dir, 'flow_dir.tif')
            pygeoprocessing.numpy_array_to_raster(
                base_array=flow_dir_array,
                target_nodata=255,
                pixel_size=(2, -2),
                origin=(2, -2),
                projection_wkt=srs,
                target_path=flow_dir_path)

            # These geometries test:
            #  * Delineation works with varying geometry types
            #  * That we exclude seed pixels that are over nodata
            #  * That we exclude seed pixels off the bounds of the raster
            horizontal_line = shapely.geometry.LineString(
                [(19, -11), (25, -11)])
            vertical_line = shapely.geometry.LineString([(21, -9), (21, -13)])
            square = shapely.geometry.box(17, -13, 21, -9)
            point = shapely.geometry.Point(21, -11)

            outflow_vector_path = os.path.join(
                self.workspace_dir, 'outflow.gpkg')
            pygeoprocessing.shapely_geometry_to_vector(
                [horizontal_line, vertical_line, square, point],
                outflow_vector_path, srs_wkt,
                'GPKG',
                {
                    'polygon_id': ogr.OFTInteger,
                    'field_string': ogr.OFTString,
                    'other': ogr.OFTReal,
                    # We use ws_id internally, so make sure that this field is
                    # copied over into the final vector since it's present in
                    # the source vector.
                    'ws_id': ogr.OFTInteger,
                },
                [
                    {'polygon_id': 1, 'field_string': 'hello world',
                     'other': 1.111, 'ws_id': 1},
                    {'polygon_id': 2, 'field_string': 'hello foo',
                     'other': 2.222, 'ws_id': 2},
                    {'polygon_id': 3, 'field_string': 'hello bar',
                     'other': 3.333, 'ws_id': 3},
                    {'polygon_id': 4, 'field_string': 'hello baz',
                     'other': 4.444, 'ws_id': 4},
                ],
                ogr_geom_type=ogr.wkbUnknown)

            target_watersheds_path = os.path.join(
                self.workspace_dir, 'watersheds.gpkg')

            pygeoprocessing.routing.delineate_watersheds_d8(
                (flow_dir_path, 1), outflow_vector_path,
                target_watersheds_path,
                target_layer_name='watersheds_something')

            try:
                watersheds_vector = gdal.OpenEx(target_watersheds_path,
                                                gdal.OF_VECTOR)
                watersheds_layer = watersheds_vector.GetLayer(
                    'watersheds_something')
                self.assertEqual(watersheds_layer.GetFeatureCount(), 4)
                self.assertEqual(watersheds_layer.GetGeomType(),
                                 ogr.wkbMultiPolygon)

                # All features should have the same watersheds, both in area
                # and geometry.
                flow_dir_bbox = pygeoprocessing.get_raster_info(
                    flow_dir_path)['bounding_box']
                expected_watershed_geometry = shapely.geometry.box(
                    *flow_dir_bbox)
                expected_watershed_geometry = (
                    expected_watershed_geometry.difference(
                        shapely.geometry.box(20, -2, 22, -10)))
                expected_watershed_geometry = (
                    expected_watershed_geometry.difference(
                        shapely.geometry.box(20, -12, 22, -22)))
                pygeoprocessing.shapely_geometry_to_vector(
                    [expected_watershed_geometry],
                    os.path.join(self.workspace_dir, 'foo.gpkg'), srs_wkt,
                    'GPKG', ogr_geom_type=ogr.wkbPolygon)

                id_to_fields = {}
                for feature in watersheds_layer:
                    geometry = feature.GetGeometryRef()
                    self.assertEqual(geometry.GetGeometryType(),
                                     ogr.wkbMultiPolygon)
                    shapely_geom = shapely.wkb.loads(
                        bytes(geometry.ExportToWkb()))
                    self.assertEqual(
                        shapely_geom.area, expected_watershed_geometry.area)
                    self.assertEqual(
                        shapely_geom.intersection(
                            expected_watershed_geometry).area,
                        expected_watershed_geometry.area)
                    self.assertEqual(
                        shapely_geom.difference(
                            expected_watershed_geometry).area, 0)

                    field_values = feature.items()
                    id_to_fields[field_values['polygon_id']] = field_values
            finally:
                watersheds_vector = None
                watersheds_layer = None

            try:
                outflow_vector = gdal.OpenEx(
                    outflow_vector_path, gdal.OF_VECTOR)
                outflow_layer = outflow_vector.GetLayer()
                found_ws_ids = set()  # make sure the ws_id field is copied
                for feature in outflow_layer:
                    self.assertEqual(
                        id_to_fields[feature.GetField('polygon_id')],
                        feature.items())
                    found_ws_ids.add(feature.GetField('ws_id'))
            finally:
                outflow_layer = None
                outflow_vector = None
            self.assertEqual(found_ws_ids, set([1, 2, 3, 4]))

    def test_split_geometry_into_seeds(self):
        """PGP watersheds: Test geometry-to-seed extraction."""
        nodata = 255
        flow_dir_array = numpy.array([
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0],
            [nodata, nodata, nodata, nodata, nodata, nodata, nodata],
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 2, 2]], dtype=numpy.uint8)
        srs = osr.SpatialReference()
        srs.ImportFromEPSG(32731)  # WGS84 / UTM zone 31s
        srs_wkt = srs.ExportToWkt()

        flow_dir_path = os.path.join(self.workspace_dir, 'flow_dir.tif')
        driver = gdal.GetDriverByName('GTiff')
        flow_dir_raster = driver.Create(
            flow_dir_path, flow_dir_array.shape[1], flow_dir_array.shape[0],
            1, gdal.GDT_Byte, options=(
                'TILED=YES', 'BIGTIFF=YES', 'COMPRESS=LZW',
                'BLOCKXSIZE=256', 'BLOCKYSIZE=256'))
        flow_dir_raster.SetProjection(srs_wkt)
        flow_dir_band = flow_dir_raster.GetRasterBand(1)
        flow_dir_band.WriteArray(flow_dir_array)
        flow_dir_band.SetNoDataValue(255)
        flow_dir_geotransform = [2, 2, 0, -2, 0, -2]
        flow_dir_raster.SetGeoTransform(flow_dir_geotransform)
        flow_dir_raster = None
        flow_dir_info = pygeoprocessing.get_raster_info(flow_dir_path)

        point = shapely.geometry.Point(2.5, -2.5)
        linestring = shapely.geometry.LineString([(10, -2), (10, -9.9)])
        box = shapely.geometry.box(4.1, -7.9, 7.9, -4.1)

        for index, (geometry, expected_seeds) in enumerate((
                (point, set([(0, 0)])),
                # includes nodata pixel
                (linestring, set([(4, 0), (4, 1), (4, 2), (4, 3)])),
                # includes nodata pixels
                (box, set([(1, 1), (2, 1), (1, 2), (2, 2)])))):

            raster_path = os.path.join(self.workspace_dir, '%s.tif' % index)
            diagnostic_path = os.path.join(
                self.workspace_dir, '%s.gpkg' % index)
            result_seeds = watershed._split_geometry_into_seeds(
                geometry.wkb, flow_dir_info['geotransform'], srs,
                flow_dir_array.shape[1], flow_dir_array.shape[0],
                raster_path, diagnostic_path)

            self.assertEqual(result_seeds, expected_seeds)

    def test_split_geometry_into_seeds_willamette(self):
        """PGP watersheds: Test geometry-to-seed extraction in Willamette."""
        flow_dir_array = numpy.array([
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0]], dtype=numpy.uint8)
        srs = osr.SpatialReference()
        srs.ImportFromEPSG(3157)  # UTM zone 17N
        srs_wkt = srs.ExportToWkt()

        flow_dir_path = os.path.join(self.workspace_dir, 'flow_dir.tif')
        driver = gdal.GetDriverByName('GTiff')
        flow_dir_raster = driver.Create(
            flow_dir_path, flow_dir_array.shape[1], flow_dir_array.shape[0],
            1, gdal.GDT_Byte, options=(
                'TILED=YES', 'BIGTIFF=YES', 'COMPRESS=LZW',
                'BLOCKXSIZE=256', 'BLOCKYSIZE=256'))
        flow_dir_raster.SetProjection(srs_wkt)
        flow_dir_band = flow_dir_raster.GetRasterBand(1)
        flow_dir_band.WriteArray(flow_dir_array)
        flow_dir_band.SetNoDataValue(255)
        pixel_xsize = 30
        pixel_ysize = -30
        flow_dir_geotransform = [
            443723, pixel_xsize, 0,
            4956546, 0, pixel_ysize]
        flow_dir_raster.SetGeoTransform(flow_dir_geotransform)
        flow_dir_raster = None
        flow_dir_info = pygeoprocessing.get_raster_info(flow_dir_path)

        pixel_indexes_array = numpy.arange(
            flow_dir_array.size).reshape(flow_dir_array.shape)
        pixel_indexes_path = os.path.join(
            self.workspace_dir, 'pixel_indexes.tif')
        driver = gdal.GetDriverByName('GTiff')
        pixel_indexes_raster = driver.Create(
            pixel_indexes_path, pixel_indexes_array.shape[1],
            pixel_indexes_array.shape[0], 1, gdal.GDT_Byte, options=(
                'TILED=YES', 'BIGTIFF=YES', 'COMPRESS=LZW',
                'BLOCKXSIZE=256', 'BLOCKYSIZE=256'))
        pixel_indexes_raster.SetProjection(srs_wkt)
        pixel_indexes_band = pixel_indexes_raster.GetRasterBand(1)
        pixel_indexes_band.WriteArray(pixel_indexes_array)
        pixel_indexes_band.SetNoDataValue(255)
        pixel_xsize = 30
        pixel_ysize = -30
        pixel_indexes_geotransform = [
            443723, pixel_xsize, 0,
            4956546, 0, pixel_ysize]
        pixel_indexes_raster.SetGeoTransform(pixel_indexes_geotransform)
        pixel_indexes_raster = None

        point = shapely.geometry.Point(
            flow_dir_geotransform[0] + pixel_xsize / 2.,
            flow_dir_geotransform[3] + pixel_ysize / 2.)
        linestring = shapely.geometry.LineString([
            (flow_dir_geotransform[0] + pixel_xsize * 4,
             # extend beyond y boundary
             flow_dir_geotransform[3] - pixel_ysize * 2),
            (flow_dir_geotransform[0] + pixel_xsize * 4,
             flow_dir_geotransform[3] + pixel_ysize * 5)])
        box = shapely.geometry.box(
            flow_dir_geotransform[0] + pixel_xsize * 2.1,
            flow_dir_geotransform[3] + pixel_ysize * 4.1,
            flow_dir_geotransform[0] + pixel_xsize * 3.9,
            flow_dir_geotransform[3] + pixel_ysize * 2.1)

        for index, (geometry, expected_seeds) in enumerate((
                (point, set([(0, 0)])),
                (linestring, set([(4, 0), (4, 1), (4, 2), (4, 3)])),
                (box, set([(2, 2), (2, 3), (3, 2), (3, 3)])))):

            result_seeds = watershed._split_geometry_into_seeds(
                geometry.wkb, flow_dir_info['geotransform'], srs,
                flow_dir_array.shape[1], flow_dir_array.shape[0],
                os.path.join(self.workspace_dir, '%s.tif' % index),
                os.path.join(self.workspace_dir, '%s.gpkg' % index))

            self.assertEqual(result_seeds, expected_seeds)



import os
import shutil
import tempfile
import unittest

import numpy
import numpy.testing
from osgeo import gdal


class KernelTests(unittest.TestCase):
    def setUp(self):
        self.workspace = tempfile.mkdtemp()
        self.filepath = os.path.join(self.workspace, 'kernel.tif')

    def tearDown(self):
        shutil.rmtree(self.workspace)

    def test_kernel_from_array(self):
        """Kernels: test kernel creation from a numpy array."""
        import pygeoprocessing
        import pygeoprocessing.kernels

        # let's use a sharpening mask, just for fun.
        # https://en.wikipedia.org/wiki/Kernel_(image_processing)
        kernel = numpy.array([
            [0, -1, 0],
            [-1, 5, -1],
            [0, -1, 0]], dtype=numpy.int32)

        pygeoprocessing.kernels.kernel_from_numpy_array(
            kernel, self.filepath)

        # This both verifies that we can read the kernel with GDAL and that
        # it's the datatype we expect.
        raster_info = pygeoprocessing.get_raster_info(self.filepath)
        self.assertEqual(raster_info['datatype'], gdal.GDT_Int32)

        numpy.testing.assert_equal(
            pygeoprocessing.raster_to_numpy_array(self.filepath),
            kernel)

    def test_kernel_invalid_dimensions(self):
        """Kernels: test error with invalid dimensions."""
        import pygeoprocessing.kernels

        kernel = numpy.array([1, 2, 3])
        with self.assertRaises(ValueError) as cm:
            pygeoprocessing.kernels.kernel_from_numpy_array(
                kernel, self.filepath)

        self.assertIn("array must have exactly 2 dimensions, not 1",
                      str(cm.exception))

    def test_dichotomy(self):
        """Kernels: test dichotomous kernel."""
        import pygeoprocessing.kernels

        max_dist = 30
        n_nonzero_pixels = 2821
        for normalize, expected_sum in [(True, 1), (False, n_nonzero_pixels)]:
            pygeoprocessing.kernels.dichotomous_kernel(
                self.filepath, max_distance=max_dist, normalize=normalize)

            kernel_array = pygeoprocessing.raster_to_numpy_array(
                self.filepath)

            self.assertEqual(kernel_array.shape, (max_dist*2+1, max_dist*2+1))
            numpy.testing.assert_allclose(kernel_array.sum(), expected_sum)
            self.assertEqual(numpy.count_nonzero(kernel_array),
                             n_nonzero_pixels)

    def test_exponential_decay(self):
        """Kernels: test exponential decay."""
        import pygeoprocessing.kernels

        max_dist = 30
        expected_distance = 15
        n_nonzero_pixels = 2821
        for normalize, expected_sum in [(True, 1), (False, 838.87195)]:
            pygeoprocessing.kernels.exponential_decay_kernel(
                self.filepath, max_distance=max_dist,
                expected_distance=expected_distance, normalize=normalize)

            kernel_array = pygeoprocessing.raster_to_numpy_array(
                self.filepath)

            self.assertEqual(kernel_array.shape, (max_dist*2+1, max_dist*2+1))
            numpy.testing.assert_allclose(kernel_array.sum(), expected_sum)
            self.assertEqual(numpy.count_nonzero(kernel_array),
                             n_nonzero_pixels)

    def test_linear_decay(self):
        """Kernels: test linear decay."""
        import pygeoprocessing.kernels

        max_dist = 30
        n_nonzero_pixels = 2809
        for normalize, expected_sum in [(True, 1), (False, 942.44055)]:
            pygeoprocessing.kernels.linear_decay_kernel(
                self.filepath, max_distance=max_dist,
                normalize=normalize)

            kernel_array = pygeoprocessing.raster_to_numpy_array(
                self.filepath)

            self.assertEqual(kernel_array.shape, (max_dist*2+1, max_dist*2+1))
            numpy.testing.assert_allclose(kernel_array.sum(), expected_sum)
            self.assertEqual(numpy.count_nonzero(kernel_array),
                             n_nonzero_pixels)

    def test_normal_distribution(self):
        """Kernels: test normal distribution."""
        import pygeoprocessing.kernels

        sigma = 3
        n_std_dev = 10
        n_nonzero_pixels = 2821
        for normalize, expected_sum in [(True, 1), (False, 0.98877025)]:
            pygeoprocessing.kernels.normal_distribution_kernel(
                self.filepath, sigma=10, n_std_dev=3,
                normalize=normalize)

            kernel_array = pygeoprocessing.raster_to_numpy_array(
                self.filepath)

            max_dist = sigma * n_std_dev
            self.assertEqual(kernel_array.shape, (max_dist*2+1, max_dist*2+1))
            numpy.testing.assert_allclose(kernel_array.sum(), expected_sum)
            self.assertEqual(numpy.count_nonzero(kernel_array),
                             n_nonzero_pixels)

    def test_create_distance_decay_kernel_callable(self):
        """Kernels: test kernel creation from function."""
        import pygeoprocessing.kernels

        # should create a sort of bullseye effect.
        # pixel outside max_distance are 0.
        # The kernel's rings alternate between -1 and 1.
        def _my_weird_kernel(dist):
            return ((numpy.floor(dist) % 2) * 2) - 1

        # do the same thing, but as a string
        my_string_kernel = "((floor(dist) % 2) * 2) - 1"

        for function in (_my_weird_kernel, my_string_kernel):
            pygeoprocessing.kernels.create_distance_decay_kernel(
                self.filepath, _my_weird_kernel, max_distance=30,
                normalize=False)

            array = pygeoprocessing.raster_to_numpy_array(self.filepath)
            self.assertEqual(array.shape, (61, 61))
            self.assertEqual(set(numpy.unique(array)), {-1, 0, 1})



"""setup.py module for PyGeoprocessing."""
import os
import platform
import subprocess

import numpy
from Cython.Build import cythonize
from setuptools import setup
from setuptools.extension import Extension

# Read in requirements.txt and populate the python readme with the non-comment
# contents.
_REQUIREMENTS = [
    x for x in open('requirements.txt').read().split('\n')
    if not x.startswith('#') and len(x) > 0]
LONG_DESCRIPTION = open('README.rst').read().format(
    requirements='\n'.join(['    ' + r for r in _REQUIREMENTS]))
LONG_DESCRIPTION += '\n' + open('HISTORY.rst').read() + '\n'

include_dirs = [
    numpy.get_include(),
    'src/pygeoprocessing/routing',
    'src/pygeoprocessing/extensions']
if platform.system() == 'Windows':
    compiler_args = ['/std:c++20']
    compiler_and_linker_args = []
    if 'PYGEOPROCESSING_GDAL_LIB_PATH' not in os.environ:
        raise RuntimeError(
            'env variable PYGEOPROCESSING_GDAL_LIB_PATH is not defined. '
            'This env variable is required when building on Windows. If '
            'using conda to manage your gdal installation, you may set '
            'PYGEOPROCESSING_GDAL_LIB_PATH=%CONDA_PREFIX%/Library".')
    library_dirs = [os.path.join(
        os.environ["PYGEOPROCESSING_GDAL_LIB_PATH"].rstrip(), "lib")]
    include_dirs.append(os.path.join(
        os.environ["PYGEOPROCESSING_GDAL_LIB_PATH"].rstrip(), "include"))
else:
    compiler_args = [subprocess.run(
        ['gdal-config', '--cflags'], capture_output=True, text=True
    ).stdout.strip()]
    compiler_and_linker_args = ['-std=c++20']
    library_dirs = [subprocess.run(
        ['gdal-config', '--libs'], capture_output=True, text=True
    ).stdout.split()[0][2:]] # get the first argument which is the library path

setup(
    name='pygeoprocessing',
    description="PyGeoprocessing: Geoprocessing routines for GIS",
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/x-rst',
    url='https://github.com/natcap/pygeoprocessing',
    packages=[
        'pygeoprocessing',
        'pygeoprocessing.routing',
        'pygeoprocessing.multiprocessing',
        'pygeoprocessing.extensions'
    ],
    package_dir={
        'pygeoprocessing': 'src/pygeoprocessing'
    },
    setup_requires=['cython', 'numpy'],
    include_package_data=True,
    install_requires=_REQUIREMENTS,
    license='BSD',
    zip_safe=False,
    ext_modules=cythonize([
        Extension(
            name="pygeoprocessing.routing.routing",
            sources=["src/pygeoprocessing/routing/routing.pyx"],
            include_dirs=include_dirs,
            library_dirs=library_dirs,
            libraries=['gdal'],
            extra_compile_args=compiler_args + compiler_and_linker_args,
            extra_link_args=compiler_and_linker_args,
            define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')],
            language="c++",
        ),
        Extension(
            "pygeoprocessing.routing.watershed",
            sources=["src/pygeoprocessing/routing/watershed.pyx"],
            include_dirs=include_dirs,
            library_dirs=library_dirs,
            libraries=['gdal'],
            extra_compile_args=compiler_args + compiler_and_linker_args,
            extra_link_args=compiler_and_linker_args,
            define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')],
            language="c++",
        ),
        Extension(
            "pygeoprocessing.geoprocessing_core",
            sources=[
                'src/pygeoprocessing/geoprocessing_core.pyx'],
            include_dirs=include_dirs,
            library_dirs=library_dirs,
            libraries=['gdal'],
            extra_compile_args=compiler_args + compiler_and_linker_args,
            extra_link_args=compiler_and_linker_args,
            define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')],
            language="c++"
        ),
    ])
)



"""A library of kernels for use in ``pygeoprocessing.convolve_2d``.

``pygeoprocessing.convolve_2d`` has some stringent requirements for its kernels
that require a thorough understanding of GDAL and GeoTiffs to be able to use,
including:

    * Kernels must be GDAL-readable rasters, preferably GeoTiffs
    * Kernels must not be striped, and, for more efficient disk I/O, should be
      tiled.
    * The pixel values of a kernel may not represent nodata (GDAL's internal
      way of indicating invalid pixel data)
    * Pixel values must be finite, real numbers.  That is, they may not
      represent positive or negative infinity or NaN.

The functions in this module represent kernels that we have found to be
commonly used, including:

    * Dichotomous kernel: :meth:`dichotomous_kernel`
    * Gaussian (normal) decay kernel: :meth:`normal_distribution_kernel`
    * Linear decay kernel: :meth:`linear_decay_kernel`
    * Exponential decay kernel: :meth:`exponential_decay_kernel`

Additionally, the user may define their own kernel using the helper functions
included here:

    * :meth:`kernel_from_numpy_array`, for kernels where the kernel array is
      already available as a ``numpy`` array, such as those used in common
      image processing operations like sharpening and edge detection.
      https://en.wikipedia.org/wiki/Kernel_(image_processing)
    * :meth:`create_distance_decay_kernel`, for kernels that are a function of
      distance from the center pixel
"""
import logging
import math
from typing import Callable
from typing import Text
from typing import Union

import numpy
import pygeoprocessing
from numpy.typing import ArrayLike
from osgeo import gdal

from .utils import gdal_use_exceptions

FLOAT32_NODATA = float(numpy.finfo(numpy.float32).min)
LOGGER = logging.getLogger(__name__)


def kernel_from_numpy_array(
        numpy_array: ArrayLike, target_kernel_path: Text) -> None:
    """Create a convolution kernel from a numpy array.

    Args:
        numpy_array: A 2-dimensional numpy array to convert into a raster.
        target_kernel_path: The path to where the kernel should be written.

    Returns:
        ``None``
    """
    n_dimensions = numpy.array(numpy_array).ndim
    if n_dimensions != 2:
        raise ValueError(
            "The kernel array must have exactly 2 dimensions, not "
            f"{n_dimensions}")

    # The kernel is technically a TIFF, not a GeoTiff.
    pygeoprocessing.numpy_array_to_raster(
        numpy_array, target_nodata=None, pixel_size=None, origin=None,
        projection_wkt=None, target_path=target_kernel_path)


def dichotomous_kernel(
        target_kernel_path: Text,
        max_distance: Union[int, float],
        normalize: bool = True) -> None:
    """Create a binary kernel indicating presence/absence within a distance.

    This is equivalent to ``int(dist <= max_distance)`` for each pixel in the
    kernel, where ``dist`` is the euclidean distance of the pixel's centerpoint
    to the centerpoint of the center pixel.

    Args:
        target_kernel_path: The path to where the kernel will be written.
            Must have a file extension of ``.tif``.
        max_distance: The distance threshold, in pixels. Kernel
            pixels that are greater than ``max_distance`` from the centerpoint
            of the kernel will have values of ``0.0``.
        normalize: Whether to normalize the kernel.

    Returns:
        ``None``
    """
    def _dichotomous(dist):
        return dist <= max_distance

    create_distance_decay_kernel(
        target_kernel_path=target_kernel_path,
        distance_decay_function=_dichotomous,
        max_distance=max_distance,
        normalize=normalize
    )


def exponential_decay_kernel(
        target_kernel_path: Text,
        max_distance: Union[int, float],
        expected_distance: Union[int, float],
        normalize: bool = True) -> None:
    """Create an exponential decay kernel.

    This is equivalent to ``e**(-dist / expected_distance)`` for each pixel in
    the kernel, where ``dist`` is the euclidean distance of the pixel's
    centerpoint to the centerpoint of the center pixel and
    ``expected_distance`` represents the distance at which the kernel's values
    reach ``1/e``.  The kernel will continue to have nonzero values out to
    ``max_distance``.

    Args:
        target_kernel_path: The path to where the kernel will be written.
            Must have a file extension of ``.tif``.
        max_distance: The maximum distance of the kernel, in pixels. Kernel
            pixels that are greater than ``max_distance`` from the centerpoint
            of the kernel will have values of ``0.0``.
        expected_distance: The distance, in pixels, from the centerpoint at
            which decayed values will equal ``1/e``.
        normalize: Whether to normalize the kernel.

    Returns:
        ``None``
    """
    def _exponential_decay(dist):
        return numpy.exp(-dist / expected_distance)

    create_distance_decay_kernel(
        target_kernel_path=target_kernel_path,
        distance_decay_function=_exponential_decay,
        max_distance=max_distance,
        normalize=normalize
    )


def linear_decay_kernel(
        target_kernel_path: Text,
        max_distance: Union[int, float],
        normalize: bool = True) -> None:
    """Create a linear decay kernel.

    This is equivalent to ``(max_distance - dist) / max_distance`` for each
    pixel in the kernel, where ``dist`` is the euclidean distance between the
    centerpoint of the current pixel and the centerpoint of the center pixel in
    the kernel.

    Args:
        target_kernel_path: The path to where the kernel will be written.
            Must have a file extension of ``.tif``.
        max_distance: The maximum distance of the kernel, in pixels. Kernel
            pixels that are greater than ``max_distance`` from the centerpoint
            of the kernel will have values of ``0.0``.
        normalize: Whether to normalize the kernel.

    Returns:
        ``None``
    """
    def _linear_decay(dist):
        return (max_distance - dist) / max_distance

    create_distance_decay_kernel(
        target_kernel_path=target_kernel_path,
        distance_decay_function=_linear_decay,
        max_distance=max_distance,
        normalize=normalize
    )


def normal_distribution_kernel(
        target_kernel_path: Text,
        sigma: Union[int, float],
        n_std_dev: Union[int, float] = 3,
        normalize: bool = True):
    """Create an decay kernel following a normal distribution.

    This is equivalent to
    ``(1/(2*pi*sigma**2))*(e**((-dist**2)/(2*sigma**2)))`` for each pixel,
    where ``dist`` is the euclidean distance between the current pixel and the
    centerpoint of the center pixel.

    Args:
        target_kernel_path: The path to where the kernel will be written.
            Must have a file extension of ``.tif``.
        sigma: The width (in pixels) of a standard deviation.
        n_std_dev: The number of standard deviations to include in the kernel.
            The kernel will have values of 0 when at a distance of
            ``(sigma * n_std_dev)`` away from the centerpoint.
        normalize: Whether to normalize the kernel.

    Returns:
        ``None``
    """
    def _normal_decay(dist):
        return (1 / (2 * numpy.pi * sigma ** 2)) * numpy.exp(
            (-dist ** 2) / (2 * sigma ** 2))

    create_distance_decay_kernel(
        target_kernel_path=target_kernel_path,
        distance_decay_function=_normal_decay,
        max_distance=(sigma * n_std_dev),
        normalize=normalize
    )


@gdal_use_exceptions
def create_distance_decay_kernel(
        target_kernel_path: str,
        distance_decay_function: Union[str, Callable],
        max_distance: Union[int, float],
        normalize=True):
    """
    Create a kernel raster based on pixel distance from the centerpoint.

    Args:
        target_kernel_path (string): The path to where the target kernel should
            be written on disk.  If this file does not have the suffix
            ``.tif``, it will be added to the filepath.
        distance_decay_function (callable or str): A python callable that takes as
            input a single 1D numpy array and returns a 1D numpy array.  The
            input array will contain float32 distances to the centerpoint pixel
            of the kernel, in units of pixels. If a ``str``, then it must be a
            python expression using the local variables:

                * ``dist`` - a 1D numpy array of distances from the
                  centerpoint.
                * ``max_dist`` - a float indicating the max distance.

        max_distance (float): The maximum distance of kernel values from
            the center point.  Values outside of this distance will be set to
            ``0.0``.
        normalize=False (bool): Whether to normalize the resulting kernel.

    Returns:
        ``None``
    """
    apothem = math.floor(max_distance)
    kernel_size = apothem * 2 + 1  # allow for a center pixel
    assert kernel_size % 2 == 1
    driver = gdal.GetDriverByName('GTiff')
    kernel_dataset = driver.Create(
        target_kernel_path.encode('utf-8'), kernel_size, kernel_size, 1,
        gdal.GDT_Float32, options=[
            'BIGTIFF=IF_SAFER', 'TILED=YES', 'BLOCKXSIZE=256',
            'BLOCKYSIZE=256'])

    # NOTE: We are deliberately NOT setting a coordinate system because it
    # isn't needed.  By omitting this, we're telling GDAL to just create a
    # TIFF.

    kernel_band = kernel_dataset.GetRasterBand(1)
    kernel_nodata = FLOAT32_NODATA
    kernel_band.SetNoDataValue(kernel_nodata)

    kernel_band = None
    kernel_dataset = None

    kernel_raster = gdal.OpenEx(target_kernel_path, gdal.GA_Update)
    kernel_band = kernel_raster.GetRasterBand(1)
    band_x_size = kernel_band.XSize
    band_y_size = kernel_band.YSize
    running_sum = 0

    # If the user provided a string rather than a callable, assume it's a
    # python expression appropriate for evaling.
    if isinstance(distance_decay_function, str):
        # Avoid recompiling on each iteration.
        code = compile(distance_decay_function, '<string>', 'eval')
        numpy_namespace = {name: getattr(numpy, name) for name in dir(numpy)}

        def distance_decay_function(d):
            result = eval(
                code,
                numpy_namespace,  # globals
                {'dist': d, 'max_dist': max_distance})  # locals
            return result

    for block_data in pygeoprocessing.iterblocks(
            (target_kernel_path, 1), offset_only=True):
        array_xmin = block_data['xoff'] - apothem
        array_xmax = min(
            array_xmin + block_data['win_xsize'],
            band_x_size - apothem)
        array_ymin = block_data['yoff'] - apothem
        array_ymax = min(
            array_ymin + block_data['win_ysize'],
            band_y_size - apothem)

        pixel_dist_from_center = numpy.hypot(
            *numpy.mgrid[
                array_ymin:array_ymax,
                array_xmin:array_xmax])

        valid_pixels = (pixel_dist_from_center <= max_distance)

        kernel = numpy.zeros(pixel_dist_from_center.shape,
                             dtype=numpy.float32)
        kernel[valid_pixels] = distance_decay_function(
            pixel_dist_from_center[valid_pixels])

        if normalize:
            running_sum += kernel.sum()

        kernel_band.WriteArray(
            kernel,
            yoff=block_data['yoff'],
            xoff=block_data['xoff'])

    kernel_raster.FlushCache()
    kernel_band = None
    kernel_raster = None

    if normalize:
        kernel_raster = gdal.OpenEx(target_kernel_path, gdal.GA_Update)
        kernel_band = kernel_raster.GetRasterBand(1)
        for block_data, kernel_block in pygeoprocessing.iterblocks(
                (target_kernel_path, 1)):
            # divide by sum to normalize
            kernel_block /= running_sum
            kernel_band.WriteArray(
                kernel_block, xoff=block_data['xoff'], yoff=block_data['yoff'])

        kernel_raster.FlushCache()
        kernel_band = None
        kernel_raster = None
